import numpy as np
import math
import copy
import matlab.engine
from scipy.stats import norm
from scipy import stats



# DECLARE THE STATIC GLOBAL VARIABLES AND INITIALIZE THE PARAMETERS
class G:
    # initialized from algorithm parameters
    total_budget = 0
    min_sample = 3
    delta = 2
    alpha = 0.90
    iseed = 0
    iseed_perturb = 0
    iteration_max = 1000
    imax = 15
    s = 2
    c = 2

    # dimension initialization from oracle parameters
    dimension = 0
    total_neighbors = 0
    optType = "minimize"
    oracle_parameter = [0,0,0]

class DecisionVariable:
    x = [[0 for i in range(G.dimension)] for j in range(G.iteration_max)]

# STORE ALL OUTPUT SEEDS ONCE GENERATED
class Dynamic:
    output_seed = []
    urv = []

# SIMULATION CALL TO ORACLE  (DO NOT CHANGE)
def oracle_call(x_input_oracle, total_sample, current_sample, problem, matlabEng):
    seed = Dynamic.output_seed[current_sample : current_sample + total_sample]
    x_input_oracle = np.array(x_input_oracle).astype(int)
    seed = np.array(seed).astype(int)
    z = matlabEng.OracleCall(matlab.double(x_input_oracle.tolist()), total_sample,
                             matlab.double(seed.tolist()), problem, nargout = 2)
    feasibility = z[0]
    if feasibility > 0 and total_sample > 1:
        oracleObservations = np.array(z[1]._data).reshape(z[1].size)[0].tolist()
    elif feasibility > 0:
        oracleObservations = [z[1]]
    else:
        oracleObservations = []
    if (G.optType == "maximize") :
        oracleObservations = [x * -1 for x in oracleObservations]
    return feasibility, oracleObservations


# RANDOM SEED GENERATOR (DO NOT CHANGE)
def u16807d():
    seed = G.iseed
    for i in range(G.total_budget):
        urv = 0
        while urv <= 0 or urv >= 1:
            temp = seed * 16807
            seed = temp % 2147483647
            urv = seed / 2147483648
        Dynamic.output_seed[i] = seed
        Dynamic.urv.append(urv)


def u16807Budgeted(budget, seed):
    randomSeed = []
    randomURV = []
    for i in range(budget):
        urv = 0
        while urv <= 0 or urv >= 1:
            temp = seed * 16807
            seed = temp % 2147483647
            urv = seed / 2147483648
        randomSeed.append(seed)
        randomURV.append(urv)
    return randomURV, randomSeed



# RANDOM SEED GENERATOR (DO NOT CHANGE)
def u16807d_randomnumber(seed):
    urv = 0
    while urv <= 0 or urv >= 1:
        temp = seed * 16807
        seed = temp % 2147483647
        urv = seed/ 2147483648
    return urv, seed


# INTEGER ORDERED SIMULATION OPTIMIZATION USING ADALINE (DO NOT CHANGE FUNCTION)
def ADALINE(x_input_opt, problem, matlabEng):
    # START MATLAB ENGINE
#    matlabEng = matlab.engine.connect_matlab()
    
    # INITIALIZE ALL THE SOLVER PARAMETERS
    escort = 1
    sol_tolerance = 0.00000005
    min_sample = G.min_sample

    # RUN THE U16807D GENERATOR 
    u16807d()

    # CREATE AND NAME THE LOG FILE THAT IS GENERATED
    filename = "ADALINE_Output" + ".txt"
    f = open(filename,'w', 1)

    # INITIALIZE ALL THE VARIABLES CORRESPONDING TO ONE RUN OF THE SIMULATION OPTIMIZATION ALGORITHM
    true_objective = []
    # iteration_Q = []
    samplesize = []
    Work = []
    Magnitude = []
    overallsample = []
    fx_iteration = []
    search_direction = []
    out_search_direction = []
    pli_sample_size = []
    ncalls = 0
    iteration = 0
    max_sample = min_sample
    DecisionVariable.x[iteration] = x_input_opt
    x_best = copy.deepcopy(DecisionVariable.x[iteration])
    f.write("Iteration\t\t"+ "   " + "Decision Variable at start\t\t" + "   "
            + "Best at end of iteration\t\t" + "   " + "Estimated Total Cost\t\t"
            + "     " + "Sample Size\t\t" + "   " + "Work\t\t"
            + "     " + "Overall Sample Size" +  "\n\n")

    feasibility = oracle_call(x_best, 1, 0, problem, matlabEng)
    if feasibility[0] == 0:
        print("Enter a feasible solution to start algorithm")
        return

    # PERFORM ADAPTIVE PLI AND NEIGHBORHOOD ENUMERATION UNTIL THE BUDGET GETS EXPENDED
    while ncalls < G.total_budget:
        # START THE ADAPTIVE PLI FRAMEWORK
        if ncalls + (G.dimension + 1) * min_sample <= G.total_budget:
            linear_interpolation = Adaptive_PLI(x_best, min_sample, ncalls, 
                                                problem, matlabEng)
            fx_pli_best = linear_interpolation[1]
            x_pli_best = linear_interpolation[0]
            search_gradient = linear_interpolation[2]
            search_sample_size = linear_interpolation[3]
            ncalls = linear_interpolation[4]
            pli_sample_size.append(search_sample_size)

            # MAKE THE SAMPLE SIZES EQUAL FOR COMPARISON PURPOSES
            estimate_best = oracle_call(x_best, search_sample_size, 0,
                                            problem, matlabEng)
            fx_best = np.mean(estimate_best[1])
            ncalls += search_sample_size

            # UPDATE X_BEST IF PLI RETURNS A BETTER SOLUTION THAN THE CURRENT BEST
            if fx_pli_best < fx_best + sol_tolerance:
                x_best = x_pli_best
                fx_best = fx_pli_best

            # PERFORM LINE SEARCH AT THE END OF THE ADAPTIVE PLI PROCEDURE
            x_temp = copy.deepcopy(x_best)
            xl = copy.deepcopy(x_temp)
            i = 0
            while i < G.imax:
                s = G.s * (G.c ** i)
                magnitude = np.linalg.norm(np.array(search_gradient))
                for k in range(len(xl)):
                    xl[k] = x_temp[k] - np.floor(search_gradient[k] / magnitude * s + np.sign(search_gradient[k])*0.5)
                feasibility_check = oracle_call(xl, 1, 0, problem, matlabEng)
                if feasibility_check[0] > 0 and ncalls + search_sample_size <= G.total_budget:
                    lineSearchObs = oracle_call(xl, search_sample_size, 0, problem, matlabEng)
                    ncalls += search_sample_size
                    sampling = np.mean(lineSearchObs[1])
                    if sampling < fx_best + sol_tolerance:
                        fx_best = sampling
                        x_best = copy.deepcopy(xl)
                        # LINE SEARCH SUCCEEDS IN FINDING A GOOD SOLUTION
                    else:
                        # LINE SEARCH FAILS DUE TO FINDING AN INFERIOR SOLUTION
                        break
                else:
                    # LINE SEARCH FAILS DUE TO FINDING AN INFEASIBLE SOLUTION
                    break
                i += 1
        elif ncalls + (G.dimension + 1) * min_sample > G.total_budget:
            # BUDGET EXCEEDED
            break
                    
        # PERFORM NEIGHBORHOOD ENUMERATION
        neighborestimation = Adaptive_NE(ncalls, x_best, search_sample_size, problem, matlabEng)
        ncalls = neighborestimation[0]
        overallsample.append(neighborestimation[1])
        samplesize.append(neighborestimation[2])
        search_direction.append(neighborestimation[3])
        fx_best = neighborestimation[6]
        linesearch_Perform = neighborestimation[10]
        if neighborestimation[5] < fx_best + sol_tolerance:
            fx_best = neighborestimation[5]
            x_best = neighborestimation[4]
        
        # CONDUCT LINE SEARCH ALONG THE SEARCH DIRECTION AFTER NEIGHBORHOOD ENUMERATION IF SAMPLING CRITERION
        # IS SATISFIED
        if ncalls < G.total_budget and search_direction[iteration] >= 0:
            max_sample_temp = max(overallsample[iteration][search_direction[iteration]],search_sample_size)
            x_temp = copy.deepcopy(x_best)
            xl = copy.deepcopy(x_temp)
            search_vector = [0] * G.dimension
            if (search_direction[iteration] % 2) == 0:
                search_vector[(search_direction[iteration] // 2)] \
                              = neighborestimation[9][search_direction[iteration]]
            else:
                search_vector[(search_direction[iteration] // 2)] = \
                              - neighborestimation[9][search_direction[iteration]]

            i = 0
            if linesearch_Perform == 1:
                while i < G.imax:
                    s = G.s * (G.c ** i)
                    magnitude = np.linalg.norm(np.array(search_vector))
                    for k in range(len(xl)):
                        xl[k] = \
                            x_temp[k] + np.floor(search_vector[k] / magnitude * s + np.sign(search_vector[k])* 0.5)
                    feasibility_check = oracle_call(xl, 1, 0, problem, matlabEng)
                    if feasibility_check[0] > 0 and \
                       ncalls + overallsample[iteration][search_direction[iteration]] <= G.total_budget:
                       lineSearchNEObs = oracle_call(xl, overallsample[iteration][search_direction[iteration]],
                                                      0, problem, matlabEng)
                       ncalls += overallsample[iteration][search_direction[iteration]]
                       sampling = np.mean(lineSearchNEObs[1])
                       if sampling < fx_best + sol_tolerance:
                           fx_best = sampling
                           x_best = copy.deepcopy(xl)
                            # NE LINE SEARCH SUCCEEDS IN FINDING A GOOD SOLUTION
                       else:
                            # NE LINE SEARCH FAILS DUE TO AN INFERIOR SOLUTION 
                            break
                    else:
                        # NE LINE SEARCH FAILS DUE TO INFEASIBILITY
                        break
                    i += 1

        # PRINT THE SEARCH DIRECTION AT THE END OF NEIGHBORHOOD ENUMERATION - MORE OF BOOK KEEPING
        if search_direction[iteration] != -1000 and search_direction[iteration] > 0:
            out_search_direction.append([0] * 1)
        else:
            out_search_direction.append([10001])
        if search_direction[iteration] > 0 and search_direction[iteration] % 2 == 0:
            out_search_direction[iteration] = search_direction[iteration] / 2 + 1
        elif search_direction[iteration] > 0 and search_direction[iteration] % 2 != 0:
            out_search_direction[iteration] = - (search_direction[iteration] + 1) / 2
        else:
            out_search_direction[iteration][0] = -1000

        # print (
        # str(iteration).rjust(20) + "   " + str(DecisionVariable.x[iteration][:]).rjust(5) + "   " + str(x_best).rjust(5) + "   " + str(
        #     fx_best).rjust(5) + "   " + str(samplesize[iteration]).rjust(5) + "   " + str(ncalls))

        f.write(str(iteration) + "  \t\t " + str([int(i) for i in DecisionVariable.x[iteration][:]])
                + "  \t\t " + str([int(i) for i in x_best]) + "   " + str(int(fx_best))
                + "  \t\t " + str(samplesize[iteration])
                + "   \t\t" + str(ncalls) + " \t\t" + str(overallsample[iteration])+  "\n\n")
        fx_iteration.append(fx_best)
        iteration += 1

        # DISCARD USED SEEDS AT THE END OF EVERY ITERATION
        max_sample = max(max_sample, max_sample_temp)
        
        tempSeeds1 = Dynamic.output_seed[max_sample:]
        tempSeeds2 = u16807Budgeted(max_sample, tempSeeds1[-1])
        
        tempSeeds1.extend(tempSeeds2[1])
        
        Dynamic.output_seed = tempSeeds1
                
#        if max_sample_temp <= max_sample:
#            max_sample_generate = u16807d_randomnumber(Dynamic.output_seed[max_sample])
#            Dynamic.output_seed[max_sample] = max_sample_generate[1]
#        else:
#            max_sample = max(max_sample,max_sample_temp)
#        Dynamic.output_seed[0] = Dynamic.output_seed[max_sample]

        # INCREMENT MINIMUM SAMPLE SIZE IF THE ESCORT SEQUENCE IS ACTIVATED
        if escort == 1: min_sample = max(int(iteration**0.65),G.min_sample)

        DecisionVariable.x[iteration][:] = x_best[:]
        Work.append(ncalls)
        Magnitude.append(np.linalg.norm
                        (np.array(DecisionVariable.x[iteration][:]) - np.array(DecisionVariable.x[iteration-1][:])))


    f.close()
    return true_objective, pli_sample_size, overallsample, ncalls, search_direction, x_best


def Adaptive_NE(ncalls, x, min_sample, problem, matlabEng):
    linesearch_NE = 0
    delta = G.delta
    incumbent_solution = x
    alpha = G.alpha
    budget_limit = 0
    diff_zone = 0

    # INITIALIZE THE PARAMETERS FOR THE NEIGHBORHOOD ENUMERATION PROCEDURE
    neighbor_solution = [[incumbent_solution[j] for j in range(G.dimension)] for i in range(G.total_neighbors)]
    individual_incumbent = []
    samplesize_incumbent = min_sample
    samplesize = [0] * G.total_neighbors
    est_value_incumbent = [0] * G.total_neighbors
    est_value = [0] * G.total_neighbors
    est_var_incumbent = 1
    est_var = [1] * G.total_neighbors
    mean_diff = [0] * G.total_neighbors
    se_diff = [1] * G.total_neighbors
    coeff_diff = [-1] * G.total_neighbors
    norm_diff = [1] * G.total_neighbors
    sum_xy = [0] * G.total_neighbors
    sum_xx = [0] * G.total_neighbors
    cov_xy = [0] * G.total_neighbors
    sample_reduced = [0] * G.total_neighbors
    sample_reduced_indices = []
    sample_only = [0] * G.total_neighbors
    sample_only_indices = []
    reduced_only = [0] * G.total_neighbors
    reduced_only_indices = []
    secondary_priority = [0] * G.total_neighbors
    secondary_priority_indices = []
    indicestosample = []
    infeasible_index = []
    x_best = [-1] * G.dimension
    fx_best = -1000

    # CREATE ALL THE CANDIDATE SOLUTIONS AND SAMPLE A MINIMUM FROM EACH CANDIDATE
    for i in range(G.total_neighbors):
        if (i % 2) == 0:
            neighbor_solution[i][i // 2] += 1
        else:
            neighbor_solution[i][(i - 1) // 2] += -1
        feasibility_check = oracle_call(neighbor_solution[i][:], 1, 0, problem, matlabEng)
        infeasible_index.append(feasibility_check[0])
    feasible_set = [i for i in range(G.total_neighbors) if infeasible_index[i] == 1]
    for i in feasible_set:
        if ncalls + min_sample <= G.total_budget:
            estimate_neighbor = oracle_call(neighbor_solution[i][:], min_sample, 0, problem, matlabEng)
            est_value[i] = np.mean(estimate_neighbor[1])
            est_var[i] = np.var(estimate_neighbor[1], ddof = 1)
            ncalls += min_sample
            samplesize[i] = min_sample
        else:
            indicestosample = [i]
            ncalls = G.total_budget + 1
            budget_limit = 1
            break
        if i == feasible_set[0]:
            estimate_incumbent = oracle_call(incumbent_solution, min_sample, 0, problem, matlabEng)
            est_value_incumbent = [np.mean(estimate_incumbent[1])] * G.total_neighbors
            est_var_incumbent = np.var(estimate_incumbent[1], ddof = 1)
            individual_incumbent.extend(estimate_incumbent[1])
            ncalls += min_sample
        mean_diff[i] = est_value_incumbent[i] - est_value[i]
        sum_xx[i] = np.dot(np.array(estimate_neighbor[1], dtype=object),np.array(estimate_neighbor[1], dtype=object))
        sum_xy[i] = np.dot(np.array(individual_incumbent, dtype=int),np.array(estimate_neighbor[1], dtype=object))
        cov_xy[i] += (sum_xy[i] - min_sample * est_value_incumbent[i] * est_value[i]) / (min_sample - 1)

        # CATEGORIZE THE FEASIBLE NEIGHBORS INTO 1 OF 4 CATEGORIES BASED ON THEIR ESTIMATED OBJECTIVE AND STD. ERROR
        if est_var[i] + est_var_incumbent - 2 * cov_xy[i] > 0:
            se_diff[i] = math.sqrt((est_var[i] + est_var_incumbent - 2 * cov_xy[i]) / min_sample)
            coeff_diff[i] = mean_diff[i] / se_diff[i]
        else:
            se_diff[i] = 0
            coeff_diff[i] = mean_diff[i] / (10**(-5))

        norm_diff[i] = 1/stats.t.ppf(alpha, min_sample-1) * abs(mean_diff[i])
        if se_diff[i] > norm_diff[i] and mean_diff[i] > diff_zone:
            reduced_only[i] = 1
            reduced_only_indices.append(i)
        elif se_diff[i] <= norm_diff[i] and mean_diff[i] > diff_zone:
            sample_reduced[i] = 1
            sample_reduced_indices.append(i)
        elif se_diff[i] <= norm_diff[i] and mean_diff[i] <= diff_zone:
            sample_only[i] = 1
            sample_only_indices.append(i)
        else:
            secondary_priority[i] = 1
            secondary_priority_indices.append(i)

    # CHECK IF ADAPTIVE NEIGHBORHOOD ENUMERATION IS REQUIRED?
    if len(sample_reduced_indices) > 0:
        linesearch_NE = 1
        list_direction = [i for i in sample_reduced_indices if mean_diff[i] > diff_zone]
    elif len(reduced_only_indices) > 0:
        list_direction = [i for i in reduced_only_indices if mean_diff[i] > diff_zone]

    # PERFORM ADAPTIVE NE PROCEDURE WHEN THERE IS NO GOOD NEIGHBORING SOLUTION TO HOP ON TO
    if len(reduced_only_indices) == 0:
        linesearch_NE = 1

        # PERFORM ADAPTIVE NE PROCEDURE
        while ncalls <= G.total_budget and budget_limit == 0 and len(sample_reduced_indices) == 0:
            samplefromincumbent = 0
            extrasample = [0] * G.total_neighbors
            if len(reduced_only_indices) + len(secondary_priority_indices) > 0:
                indicestosample = reduced_only_indices[:] + secondary_priority_indices[:]
                rcondition = 1
            elif len(sample_only_indices) == len(feasible_set):
                indicestosample = sample_only_indices[:]
                rcondition = 3
            templist = indicestosample[:]
            for j in templist:
                extrasample[j] = samplesize_incumbent - samplesize[j]
                if extrasample[j] == 0:
                    indicestosample.remove(j)
            if max(extrasample) == 0:
                indicestosample = templist[:]
                for k in indicestosample:
                    extrasample[k] = delta
                samplefromincumbent = 1
            if samplefromincumbent == 1:
                new_sample_incumbent = oracle_call(incumbent_solution, delta, samplesize_incumbent, problem, matlabEng)
                individual_incumbent.extend(new_sample_incumbent[1])
                samplesize_incumbent += delta
                ncalls += delta
                current_estimate = np.mean(individual_incumbent)

    # SAMPLE ALL THE NEIGHBORS THAT DO NOT SATISFY THE SAMPLING CRITERION
            for j in indicestosample:
                if ncalls + extrasample[j] < G.total_budget:
                    new_sample_neighbor = oracle_call(neighbor_solution[j][:], extrasample[j], samplesize[j],
                                                      problem, matlabEng)
                else:
                    ncalls = G.total_budget + 1
                    budget_limit = 1
                    break
                sum_xx[j] += \
                    np.dot(np.array(new_sample_neighbor[1], dtype=object), np.array(new_sample_neighbor[1], dtype=object))
                sum_xy[j] += \
                    np.dot(np.array(new_sample_neighbor[1], dtype=object), np.array(individual_incumbent[-extrasample[j]:], dtype=int))
                est_value[j] = (est_value[j] * samplesize[j] + np.mean(new_sample_neighbor[1]) * extrasample[j]) / (samplesize[j] + extrasample[j])
                est_var[j] = (sum_xx[j] - (samplesize[j] + extrasample[j]) * est_value[j] ** 2) / (
                (samplesize[j] + extrasample[j]) - 1)
                ncalls += extrasample[j]
                samplesize[j] += extrasample[j]
                est_value_incumbent[j] = np.mean(individual_incumbent[0:samplesize[j]])
                est_var_incumbent = np.var(individual_incumbent, ddof=1)
                cov_xy[j] = (sum_xy[j] - samplesize[j] * est_value[j] * est_value_incumbent[j]) / (samplesize[j] - 1)
                mean_diff[j] = est_value_incumbent[j] - est_value[j]
                if est_var[j] + est_var_incumbent - 2 * cov_xy[j] > 0:
                    se_diff[j] = math.sqrt((est_var[j] + est_var_incumbent - 2 * cov_xy[j]) / samplesize_incumbent)
                else:
                    se_diff[j] = 0
                coeff_diff[j] = mean_diff[j] / se_diff[j]
                norm_diff[j] = 1/stats.t.ppf(alpha,samplesize_incumbent-1) * abs(mean_diff[j])

                # UPDATE THE NEIGHBORS BASED ON THE NEW INFORMATION AFTER SAMPLING
                if se_diff[j] <= norm_diff[j] and mean_diff[j] > diff_zone:
                    sample_reduced[j] = 1
                    sample_reduced_indices.append(j)
                    if se_diff[j] == 0:
                        coeff_diff[j] = mean_diff[j] / (10**(-7))
                elif se_diff[j] <= norm_diff[j] and mean_diff[j] <= diff_zone and rcondition != 3:
                    sample_only[j] = 1
                    sample_only_indices.append(j)
                elif se_diff[j] > norm_diff[j] and rcondition != 3:
                    if mean_diff[j] <= diff_zone and j in reduced_only_indices:
                        secondary_priority[j] = 1
                        secondary_priority_indices.append(j)
                        reduced_only[j] = 0
                        reduced_only_indices.remove(j)
                    elif mean_diff[j] > diff_zone and j in secondary_priority_indices:
                        reduced_only[j] = 1
                        reduced_only_indices.append(j)
                        secondary_priority[j] = 0
                        secondary_priority_indices.remove(j)
                if se_diff[j] <= norm_diff[j] and rcondition != 3:
                    if j in reduced_only_indices:
                        reduced_only[j] = 0
                        reduced_only_indices.remove(j)
                    elif j in secondary_priority_indices:
                        secondary_priority[j] = 0
                        secondary_priority_indices.remove(j)
                if rcondition == 3:
                    if se_diff[j] > norm_diff[j] and mean_diff[j] <= diff_zone:
                        secondary_priority[j] = 1
                        secondary_priority_indices.append(j)
                        sample_only[j] = 0
                        sample_only_indices.remove(j)
                    elif se_diff[j] > norm_diff[j] and mean_diff[j] > diff_zone:
                        reduced_only[j] = 1
                        reduced_only_indices.append(j)
                        sample_only[j] = 0
                        sample_only_indices.remove(j)
                    elif se_diff[j] <= norm_diff[j] and mean_diff[j] > diff_zone:
                        sample_only[j] = 0
                        sample_only_indices.remove(j)
            #print current_estimate, est_value_incumbent
            for k in sample_only_indices:
                if current_estimate - est_value[k] > diff_zone :
                    sample_only[k] = 0
                    sample_only_indices.remove(k)
                    reduced_only[k] = 1
                    reduced_only_indices.append(k)
        list_direction = [i for i in sample_reduced_indices if coeff_diff[i] > 0]
    max_sample = np.argmax(samplesize)

    # UPDATE X_BEST AND FX_BEST
    if len(list_direction) > 0:
        best_direction = list_direction[np.argmax([mean_diff[i] for i in list_direction])]
        x_best = neighbor_solution[best_direction]
        fx_best = est_value[best_direction]
        search_direction = best_direction
        # print fx_best
    elif ncalls > G.total_budget:
        budget_limit = 1
        search_direction = -1000
        fx_best = min(est_value_incumbent[indicestosample[0]],min(est_value))
    else:
        search_direction = -1
        x_best = incumbent_solution
    # print search_direction, samplesize
    # print incumbent_solution, x_best, "\n"
    return ncalls, samplesize, samplesize_incumbent, search_direction, x_best, fx_best, est_value_incumbent[
        max_sample], neighbor_solution, budget_limit, mean_diff, linesearch_NE


def Adaptive_PLI(x_in,sample_size,ncalls, problem, matlabEng):
    delta = G.delta
    alpha = G.alpha
    x_best = x_in
    infeasible_simplex = 0
    d = G.dimension
    x = [[0 for i in range(d)] for j in range(d+1)]
    grad = [0] * d
    var_grad = [0] * d
    cov_grad = [0] * d
    sum_xy = [0] * d
    sum_xx = [0] * (d+1)
    w = [0] * (d+1)
    fnx = [100000000000000] * (d+1)
    varfnx = [0] * (d+1)
    npoints = 0
    fbar = 0
    
    seed = u16807d_randomnumber(G.iseed_perturb)
    np.random.seed(G.iseed_perturb)
    perturb_urv = np.random.uniform(0,1,d)
    G.iseed_perturb = seed[1]

    pert = [x_in[i] + 0.3*(perturb_urv[i]-0.5) for i in range(d)]
    
    x[0] = np.floor(pert)
    z = pert - x[0]
    z = np.insert(z,0,1)
    z = np.insert(z,len(z),0)
    p = np.argsort(z)
    for i in range(1,d+1):
        x[i][:] = x[i-1][:]
        x[i][p[len(p)-i-1]-1] += 1
        w[i-1] = z[p[len(p)-i]] - z[p[len(p)-i-1]]
    w[d] = z[p[1]] - z[p[0]]
    if (x_in in x[i] for i in range(d+1)):
        npoints -= 1
    # SAMPLE FROM EVERY POINT IN THE (d+1) SIMPLEX
    for i in range(d+1):
        feasibility_check = oracle_call(x[i], 1, 0, problem, matlabEng)
        if feasibility_check[0] == 1:
            npoints += 1
            estimate_x = oracle_call(x[i], sample_size, 0, problem, matlabEng)
            fnx[i] = np.mean(estimate_x[1])
            sum_xx[i] = np.dot(np.array(estimate_x[1], dtype=object), np.array(estimate_x[1], dtype=object))
            varfnx[i] = (sum_xx[i] - (sample_size * fnx[i] ** 2)) / (sample_size - 1)
            fbar += w[i] * fnx[i]
            if i > 0:
                sum_xy[i-1] = np.dot(np.array(estimate_x[1], dtype=object), np.array(old_collection, dtype=object))
                cov_grad[i-1] = (sum_xy[i-1] - sample_size * fnx[i] * fnx[i-1])/(sample_size-1)
                grad[p[len(p)-i-1]-1] = fnx[i] - ghatold
                var_grad[p[len(p)-i-1]-1] = varfnx[i] + varghatold - 2 * cov_grad[i-1]
            ghatold = fnx[i]
            varghatold = varfnx[i]
            old_collection = estimate_x[1]
        else:
            npoints = 0
            grad = [0] * d
            fx_pli_best = 1000000000
            ghatold = 0
            infeasible_simplex = 1
            break
            
    se_grad = math.sqrt(np.linalg.norm(var_grad,1))/math.sqrt(sample_size)
    norm_grad = 1/stats.t.ppf(alpha,sample_size-1) * np.linalg.norm(grad)
    ncalls += npoints * sample_size

    # ADAPTIVELY SAMPLE UNTIL THE SAMPLING CRITERION IS SATISFIED
    while se_grad > norm_grad and ncalls <= G.total_budget + 1 and infeasible_simplex == 0:
        for i in range(d+1):
            new_x = oracle_call(x[i], delta, sample_size, problem, matlabEng)
            fnx[i] = (fnx[i] * sample_size + np.sum(new_x[1])) / (sample_size + delta)
            sum_xx[i] += np.dot(np.array(new_x[1], dtype=object), np.array(new_x[1], dtype=object))
            varfnx[i] = (sum_xx[i] - ((sample_size + delta) * fnx[i] ** 2)) / (sample_size + delta - 1)
            fbar += w[i] * fnx[i]
            if i > 0:
                sum_xy[i-1] += np.dot(np.array(new_x[1], dtype=object), np.array(old_collection, dtype=object))
                cov_grad[i-1] = (sum_xy[i-1] - (sample_size + delta) * fnx[i] * fnx[i-1])/(sample_size + delta -1)
                grad[p[len(p)-i-1]-1] = fnx[i] - ghatold
                var_grad[p[len(p)-i-1]-1] = varfnx[i] + varghatold - 2 * cov_grad[i-1]
            ghatold = fnx[i]
            varghatold = varfnx[i]
            old_collection = new_x[1]
        sample_size += delta
        se_grad = math.sqrt(np.linalg.norm(var_grad,1))/math.sqrt(sample_size)
        norm_grad = 1/stats.t.ppf(alpha,sample_size-1) * np.linalg.norm(grad)
        ncalls += npoints * delta

    if infeasible_simplex == 0:
        fx_pli_best = np.min(fnx)
        x_best = x[np.argmin([fnx[i] for i in range(d+1)])]

    return x_best, fx_pli_best, grad, sample_size, ncalls



